SETTING_RELATIVE_MOVEMENT = "savegame.mod.telekinesis.relative_movement"
SETTING_REMOVE_UNBREAKABLE_ON_DISMOUNT = "savegame.mod.telekinesis.remove_unbreakable"
SETTING_CAMERA_KEEP_OBJECT_VISIBLE_BEHIND_WALLS = "savegame.mod.telekinesis.camera_raytrace"

relative_movement = GetBool(SETTING_RELATIVE_MOVEMENT)
remove_unbreakable = GetBool(SETTING_REMOVE_UNBREAKABLE_ON_DISMOUNT)
camera_raytrace = GetBool(SETTING_CAMERA_KEEP_OBJECT_VISIBLE_BEHIND_WALLS)

function init()
	RegisterTool("control_body", "Telekinesis", nil)
	SetBool("game.tool.control_body.enabled", true)
	SetFloat("game.tool.control_body.ammo", 10000)
	
	if not HasKey(SETTING_CAMERA_KEEP_OBJECT_VISIBLE_BEHIND_WALLS) then
		SetBool(SETTING_CAMERA_KEEP_OBJECT_VISIBLE_BEHIND_WALLS, true)
		camera_raytrace = true
	end
end

local function getPlayerLookPos(maxDist)
	local tf = GetPlayerCameraTransform()
	local pos = tf.pos
	local dir = TransformToParentVec(tf, Vec(0, 0, -1))

	local hit, dist, normal, shape = QueryRaycast(pos, dir, maxDist  or 1000)
	if hit then
		local hitPos = VecAdd(pos, VecScale(dir, dist))
		return hitPos, hit, dist, normal, shape
	end
end

controlBody = nil
controlBodyTime = 0
startTf = nil

startBodyTf = nil
startCamTf = nil

fpcTf = nil

size = 0
sizeLerp = 0
zoom = 0
zoomLerp = 0

lockedRotation = nil

laserDirections = {}


toggleables = {
	unbreakable = false,
	spin = false,
	zerog = false,
	light = false,
	autodig = false,
	fpc = false,
	push = false,
	pull = false,
	fire = false,
	laser = false,
	rotationlock = false,
}

function unmount()
	--toggleables.unbreakable = HasTag(controlBody, "unbreakable")
	if not toggleables.fpc then
		SetPlayerTransform(startTf, true)
	end
	
	if remove_unbreakable then
		RemoveTag(controlBody, "unbreakable")
	end
	controlBody = nil
	controlBodyTime = 0
	startTf = nil
	fpcTf = nil
end
function mount(body)
	controlBodyTime = 0
	controlBody = body
	if toggleables.unbreakable then
		SetTag(controlBody, "unbreakable")
	elseif not HasTag(controlBody, "unbreakable") then
		if toggleables.autodig then
			toggleables.autodig = false -- prevent unlucky
		end
		if toggleables.laser then
			toggleables.laser = false -- prevent another unlucky
		end
	end
	if toggleables.rotationlock then
		toggleables.rotationlock = false
	end
	
	
	local min, max = GetBodyBounds(body)
	local boundsSize = VecSub(max, min)
	size = math.max(boundsSize[1], boundsSize[2], boundsSize[3])
	
	startTf = GetPlayerTransform(true)
end
function DebugAABB(min, max)
	-- top
	DebugLine(Vec(max[1], max[2], min[3]), Vec(max[1], max[2], max[3]), 1, 1, 0, 1)
	DebugLine(Vec(min[1], max[2], max[3]), Vec(max[1], max[2], max[3]), 1, 1, 0, 1)
	DebugLine(Vec(min[1], max[2], min[3]), Vec(min[1], max[2], max[3]), 1, 1, 0, 1)
	DebugLine(Vec(min[1], max[2], min[3]), Vec(max[1], max[2], min[3]), 1, 1, 0, 1)
	
	-- bottom
	DebugLine(Vec(max[1], min[2], min[3]), Vec(max[1], min[2], max[3]), 1, 1, 0, 1)
	DebugLine(Vec(min[1], min[2], max[3]), Vec(max[1], min[2], max[3]), 1, 1, 0, 1)
	DebugLine(Vec(min[1], min[2], min[3]), Vec(min[1], min[2], max[3]), 1, 1, 0, 1)
	DebugLine(Vec(min[1], min[2], min[3]), Vec(max[1], min[2], min[3]), 1, 1, 0, 1)
	
	-- sides
	DebugLine(Vec(max[1], min[2], max[3]), Vec(max[1], max[2], max[3]), 1, 1, 0, 1)
	DebugLine(Vec(min[1], min[2], max[3]), Vec(min[1], max[2], max[3]), 1, 1, 0, 1)
	DebugLine(Vec(max[1], min[2], min[3]), Vec(max[1], max[2], min[3]), 1, 1, 0, 1)
	DebugLine(Vec(min[1], min[2], min[3]), Vec(min[1], max[2], min[3]), 1, 1, 0, 1)
end
function updateMount()
	if not controlBody then return end
	SetString("game.player.tool", "control_body")
	SetBool("game.input.locktool", true)
	SetPlayerVelocity(Vec())
	if InputPressed("r") or not IsHandleValid(controlBody) then
		unmount()
		return true
	end
	local body = controlBody
	local now = GetTime()
	local scroll = InputValue("mousewheel")
	if scroll then
		zoomLerp = zoomLerp + scroll
	end
	zoom = zoom + (zoomLerp - zoom) * 0.1
	
	local bodyTf = GetBodyTransform(body)
	local com = GetBodyCenterOfMass(body)
	local centerOfMassPos = TransformToParentPoint(bodyTf, com)
	local center = centerOfMassPos
	
	--local min, max = GetBodyBounds(body)
	--local boundsSize = VecSub(max, min)
	--local center = VecLerp(min, max, 0.5)
	--DebugPrint(VecLength(VecSub(worldPos, center)))
	--DebugCross(worldPos, 1, 1, 0, 1)
	--DebugCross(center, 1, 0, 0, 1)
	
	-- first tick when rmb is pressed, dont use camera so the head bob tilt is not in the startCamTf
	local cameraTf = (InputDown("rmb") and not InputPressed("rmb")) and GetPlayerCameraTransform() or GetPlayerTransform(true)
	
	if InputPressed("rmb") then
		startBodyTf = TransformCopy(bodyTf)
		startCamTf = TransformCopy(cameraTf)

		-- prevent prop from snapping to camera when activated
		local dir = TransformToParentVec(bodyTf, Vec(0, 0, -1))
		local rot = QuatLookAt(Vec(), dir)
		cameraTf.rot = rot
		--local rotTf = TransformToParentTransform(cameraTf, startBodyTf)
		--cameraTf.rot = rotTf.rot
	end
	if InputDown("rmb") and startCamTf then
		--local dir = TransformToParentVec(bodyTf, Vec(0, 0, -1))
		--DebugLine(startBodyTf.pos, VecAdd(startBodyTf.pos, dir), 1, 1, 1, 1)
		
		local dir = TransformToParentVec(cameraTf, Vec(0, 0, -1))
		local rot = QuatLookAt(Vec(), dir)
		bodyTf.rot = rot
		if toggleables.rotationlock then
			lockedRotation = rot
		end
		SetBodyTransform(body, Transform(startBodyTf.pos, bodyTf.rot))
		
		--local rotTf = TransformToParentTransform(startBodyTf, cameraTf)
		--rotTf.pos = startBodyTf.pos
		--SetBodyTransform(body, rotTf)
		SetBodyVelocity(body, Vec())
		SetBodyAngularVelocity(body, Vec())
	end
	if InputReleased("rmb") and startCamTf then
		cameraTf = startCamTf
		startBodyTf = nil
		startCamTf = nil
	end
	
	SetPlayerTransform(Transform(startTf.pos, cameraTf.rot), true)
	
	if InputDown("rmb") then
		cameraTf.rot = startCamTf.rot
	end
	
	local min, max = GetBodyBounds(body)
	local boundsSize = VecSub(max, min)
	--local centerBoundsPos = VecLerp(min, max, 0.5)
	--DebugCross(centerBoundsPos)
	
	sizeLerp = math.max(boundsSize[1], boundsSize[2], boundsSize[3])
	size = size + (sizeLerp - size) * 0.01
	local scaleFactor = math.max(0, size - zoom)
	
	--DebugAABB(min, max)
	
	if InputPressed("u") then
		toggleables.unbreakable = not toggleables.unbreakable
		if toggleables.unbreakable then
			SetTag(body, "unbreakable")
		else
			RemoveTag(body, "unbreakable")
		end
		
	end
	if InputPressed("l") then
		if InputDown("shift") then
			toggleables.laser = not toggleables.laser
			if toggleables.laser then
				laserDirections = {}
				local lasers = 3 + math.floor(math.random() * 5)
				for i = 1, lasers do
					table.insert(laserDirections, Vec(1 - math.random() * 2, 1 - math.random() * 2, 1 - math.random() * 2))
				end
			end
		else
			toggleables.light = not toggleables.light
		end
	end
	if InputPressed("e") then
		Explosion(center, 1)
	end
	if InputPressed("g") then
		toggleables.autodig = not toggleables.autodig
	end
	if InputPressed("p") then
		toggleables.pull = not toggleables.pull
	end
	if InputPressed("o") then
		toggleables.push = not toggleables.push
	end
	if InputPressed("c") then
		toggleables.fpc = not toggleables.fpc
		if toggleables.fpc then
			fpcTf = GetPlayerTransform(true)
			SetPlayerTransform(startTf, true)
		elseif fpcTf then
			startTf = GetPlayerTransform(true)
			SetPlayerTransform(fpcTf, true)
		end
	end
	if InputPressed("v") then
		toggleables.fire = not toggleables.fire
	end
	if InputPressed("z") then
		toggleables.zerog = not toggleables.zerog
	end
	if InputPressed("i") then
		toggleables.spin = not toggleables.spin
		if toggleables.spin then
			toggleables.rotationlock = false
		end
	end
	if InputDown("i") then
		local dir = TransformToParentVec(cameraTf, Vec(0, 0, -1))
		SetBodyAngularVelocity(body, VecAdd(GetBodyAngularVelocity(body), dir))
	end
	if InputPressed("j") then
		local delete = InputDown("shift")
		local shapes = GetBodyShapes(body)
		for i = 1, #shapes do
			local s = shapes[i]
			local joints = GetShapeJoints(s)
			for k = 1, #joints do
				local j = joints[k]
				local otherShape = GetJointOtherShape(j, s)
				if GetShapeBody(otherShape) ~= body then
					if delete then
						Delete(j)
					else
						DetachJointFromShape(j, otherShape)
					end
				end
			end
		end
		SetBodyVelocity(body, GetBodyVelocity(body)) -- force object update
	end
	if InputPressed("m") then
		toggleables.rotationlock = not toggleables.rotationlock
		if toggleables.rotationlock then
			lockedRotation = bodyTf.rot
			toggleables.spin = false
		end
	end
	if toggleables.spin then
		--local dir = TransformToParentVec(cameraTf, Vec(0, 0, -1)) -- spin positive x (right) relative to camera
		local dir = TransformToParentVec(cameraTf, Vec(-1, 0, 0)) -- spin positive z (away) relative to camera
		SetBodyAngularVelocity(body, VecScale(dir, 20))
	end
	if toggleables.light then
		PointLight(center, 1, 1, 1, size)
	end
	
	if toggleables.autodig then
		MakeHole(center, size, size, size)
	end
	if toggleables.laser then
		for i = 1, #laserDirections do
			local dir = TransformToParentVec(Transform(center, bodyTf.rot), laserDirections[i])
			dir = VecNormalize(dir)
			--DebugWatch("laserDirections[" .. i .. "]", laserDirections[i])
			--DebugWatch("dir[" .. i .. "]", dir)
			QueryRejectBody(body)
			--QueryRequire("physical")
			local hit, dist, normal, shape = QueryRaycast(center, dir, 1000)
			if hit then
				local hitPos = VecAdd(center, VecScale(dir, dist))
				Explosion(hitPos, 0.5)
			end
			DrawLine(center, VecAdd(center, VecScale(dir, hit and dist or size * 1.2)), hit and 0 or 1, hit and 1 or 0, 0, 1)
		end
	end
	if toggleables.push or toggleables.pull then
		local maxDist = size * 5
		local minPushPull = VecSub(center, Vec(maxDist, maxDist, maxDist))
		local maxPushPull = VecAdd(center, Vec(maxDist, maxDist, maxDist))
		--DebugAABB(minPushPull, maxPushPull)
		QueryRequire("physical dynamic")
		QueryRejectBody(body)
		local bodies = QueryAabbBodies(minPushPull, maxPushPull)
		for i = 1, #bodies do
			local b = bodies[i]
			local bpos = GetBodyTransform(b)
			local bcom = GetBodyCenterOfMass(b)
			local bwp = TransformToParentPoint(bpos, bcom)
			local bcenter = bwp
			local bmass = GetBodyMass(b)
			local dist = VecLength(VecSub(center, bcenter))
			if dist < maxDist then
				local vehicle = GetBodyVehicle(b)
				if vehicle and vehicle > 0 then
					local health = GetVehicleHealth(vehicle)
					DriveVehicle(vehicle, 0, 0, 0, false) -- force vehicle update to update vehicle health
					if health == 0 then
						Delete(vehicle)
					end
				end
				--DrawBodyOutline(b, 1, 0, 0, 1)
				RemoveTag(b, "unbreakable")
				if toggleables.pull then
					ApplyBodyImpulse(b, bwp, VecScale(VecNormalize(VecSub(center, bcenter)), bmass + (maxDist / dist) * 0.1))
				end
				if toggleables.push then
					ApplyBodyImpulse(b, bwp, VecScale(VecNormalize(VecSub(bcenter, center)), bmass + (maxDist / dist) * 0.1))
				end
			end
		end
	end
	if toggleables.fire then
		local step = .25
		if size >= 10 then step = 1 end
		for x = min[1], max[1], step do
			for y = min[2], max[2], step do
				for z = min[3], max[3], step do
					SpawnFire(Vec(x, y, z))
				end
			end
		end
	end
	
	if toggleables.rotationlock then
		SetBodyTransform(body, Transform(bodyTf.pos, lockedRotation))
		SetBodyAngularVelocity(body, Vec())
		--SetBodyAngularVelocity(body, VecScale(GetBodyAngularVelocity(body), -1))
	end
	--if true then
	--	QueryRejectBody(body)
	--	local hit, point = QueryClosestPoint(bodyTf.pos, size)
	--	if hit then
	--		ApplyBodyImpulse(body, center, VecScale(VecSub(bodyTf.pos, point), GetBodyMass(body)))
	--	end
	--	
	--end
	
	local tf = TransformCopy(bodyTf)
	if relative_movement then
		tf.rot = cameraTf.rot
	else
		local x, y, z = GetQuatEuler(cameraTf.rot)
		x = 0
		tf.rot = QuatEuler(x, y, z)
	end
	
	-- steering of body
	local mass = GetBodyMass(body)
	local steering, vertical = false, false
	if InputDown("up") then
		local dir = TransformToParentVec(tf, Vec(0, 0, -1))
		ApplyBodyImpulse(body, center, VecScale(dir, mass))
		steering = true
	end
	if InputDown("down") then
		local dir = TransformToParentVec(tf, Vec(0, 0, 1))
		ApplyBodyImpulse(body, center, VecScale(dir, mass))
		steering = true
	end
	if InputDown("left") then
		local dir = TransformToParentVec(tf, Vec(-1, 0, 0))
		ApplyBodyImpulse(body, center, VecScale(dir, mass))
		steering = true
	end
	if InputDown("right") then
		local dir = TransformToParentVec(tf, Vec(1, 0, 0))
		ApplyBodyImpulse(body, center, VecScale(dir, mass))
		steering = true
	end
	if InputDown("jump") then
		local dir = TransformToParentVec(tf, Vec(0, 1, 0))
		ApplyBodyImpulse(body, center, VecScale(dir, mass))
		steering = true
		vertical = true
	end
	if InputDown("crouch") then
		local dir = TransformToParentVec(tf, Vec(0, -1, 0))
		ApplyBodyImpulse(body, center, VecScale(dir, mass))
		steering = true
		vertical = true
	end
	if steering then DrawBodyOutline(body, 1, 1, 1, 1) end
	
	if not vertical and toggleables.zerog then
		local vel = GetBodyVelocity(body)
		SetBodyVelocity(body, Vec(vel[1], 0.1682, vel[3]))
	end
	
	SetPlayerHealth(1) -- dying is lame
	
	if not relative_movement then
		tf.rot = cameraTf.rot
	end
	
	if not toggleables.fpc then
		local reverseCamDir = TransformToParentVec(cameraTf, Vec(0, 0, 1))
		tf.pos = VecAdd(center, VecScale(reverseCamDir, scaleFactor))
		
		if camera_raytrace then
			QueryRejectBody(body)
			QueryRequire("physical static large")
			local hit, dist, normal, shape = QueryRaycast(center, reverseCamDir, scaleFactor)
			if hit then
				local hitPoint = VecAdd(VecAdd(center, VecScale(reverseCamDir, dist)), VecScale(reverseCamDir, -0.25))
				tf.pos = hitPoint
			end
		end
		SetCameraTransform(tf)
	else
		if InputDown("rmb") then
			SetCameraTransform(cameraTf)
		--else -- follow cam
		--	cameraTf.pos = GetPlayerCameraTransform().pos
		--	cameraTf.rot = QuatLookAt(cameraTf.pos, center)
		--	SetCameraTransform(cameraTf)
		end
	end
	return true
end

local options = {
	{
		key = "R",
		name = "Dismount"
	},
	{
		key = "RMB",
		name = "Rotate"
	},
	{
		key = "J",
		name = "Detach Joints"
	},
	{
		key = "I",
		name = "Spin",
		var = "spin"
	},
	{
		key = "Z",
		name = "Zero G",
		var = "zerog"
	},
	{
		key = "E",
		name = "Explode"
	},
	{
		key = "U",
		name = "Unbreakable",
		var = "unbreakable"
	},
	{
		key = "L",
		name = "Light",
		var = "light"
	},
	{
		key = "G",
		name = "Autodig",
		var = "autodig"
	},
	{
		key = "C",
		name = "Toggle POV",
		var = "fpc"
	},
	{
		key = "O",
		name = "Push",
		var = "push"
	},
	{
		key = "P",
		name = "Pull",
		var = "pull"
	},
	{
		key = "V",
		name = "Fire",
		var = "fire"
	},
	{
		key = "Shift + L",
		name = "Laser",
		var = "laser"
	},
	{
		key = "M",
		name = "Lock Rotation",
		var = "rotationlock"
	}
}
function draw(dt)
	if controlBody then
		controlBodyTime = math.min(controlBodyTime + dt, 1)
		UiPush()
		UiTranslate(UiWidth(), UiHeight())
		
		local ox, oy = 20, 20
		local bw, bh = 10, 10
		
		local wkey, hkey, wname, hname, hs = 0, 0, 0, 0, oy
		for i = 1, #options do
			local opt = options[i]
			UiFont("bold.ttf", 25)
			local wkey1, hkey1 = UiGetTextSize(opt.key)
			UiFont("regular.ttf", 25)
			local wname1, hname1 = UiGetTextSize(opt.name)
			wkey = math.max(wkey, wkey1)
			hkey = math.max(hkey1, hkey1)
			wname = math.max(wname, wname1)
			hname = math.max(hname, hname1)
		end
		
		local w = wkey + wname + ox * 2
		local maxtextheight = math.max(hkey, hname)
		hs = hs + maxtextheight * #options
		
		UiColor(0, 0, 0, 0.5)
		UiTranslate(-w, -hs)
		UiTranslate(-ox, -oy)
		UiImageBox("ui/common/optionsbox.png", w, hs, bw, bh)
		
		UiTranslate(ox, oy)
		UiTranslate(0, bw)
		
		UiColor(1, 1, 1, controlBodyTime)
		
		for _, opt in ipairs(options) do
			UiPush()
				UiPush()
					UiFont("bold.ttf", 25)
					UiTranslate(wkey - (UiGetTextSize(opt.key)))
					UiText(opt.key)
				UiPop()
				UiTranslate(wkey + 10)
				UiFont("regular.ttf", 25)
				if opt.var and toggleables[opt.var] then
					UiColor(0.8, 1, 0.8, controlBodyTime)
				end
				UiText(opt.name)
			UiPop()
			UiTranslate(0, maxtextheight)
		end
		
		UiPop()
	end
end

function tick(dt)
	--DebugPrint(dt)
	if updateMount() then return end
	
	if not GetBool("game.player.canusetool") then
		return
	end
	local tool = GetString("game.player.tool")
	if tool == "control_body" then
		QueryRequire("physical dynamic")
		local hitPos, hit, dist, normal, shape = getPlayerLookPos()
		if hit then
			local body = GetShapeBody(shape)
			DrawBodyOutline(body, 1, 1, 1, 1)
			if InputPressed("usetool") or InputPressed("r") then
				mount(body)
			end
		end
	end
end